// Copyright (c) 2004 Lyndon Tremblay. All Right Reserved.
#include <Main.h>

@implementation STObject

#if 1
+ (STObject *)proxy:(union STProxy *)proxy
			  class:(Class)class
			  using:(OOP)object
{
	//Class c = objc_lookup_class([ST classNameOf: object]);

	if (!class)
	{
		printf("Failed proxy.\n");
		return nil;
	}

	memset(proxy, 0, sizeof(*proxy));
	STObject *obj = (STObject *)proxy->buf;

	obj->isa = class;
	obj->oop = object;

	return obj;
}

+ alloc
{
	return class_create_instance(self);
}

- free
{
	return object_dispose(self);
}

#else

+ __new
{
	//STObject *s = [self basicAlloc];
	//[s oop: VMp->objectAlloc([self asOop], 4)];
	return [ST allocObjectForOOP: [self asOOP]];
}

+ (OOP)asOop
{
	return VMp->classNameToOOP(((Class)self)->name);
}

+ basicAlloc
{
	OOP o = VMp->objectAlloc([self asOop], 16);
	//OOP o = instantiate([self asOop], 4);

	//id new = o + sizeof(gst_object_header);
	//o->isa = [ST classOf: o];

	//((struct objc_object *)o)->class_pointer = [ST classOf: o];
	((mst_Object)o)->isa = [ST classOf: o];
	return (id)o;
	//return class_create_instance(self);
}

- (OOP)asOop
{
	//return (OOP)((char *)self + (sizeof(Class)));

	//return (OOP)(((char *)self) - sizeof(mst_Object));

	// eww >_<
	return (OOP)(((int *)self)[-2]);
}

- free
{
	//return object_dispose(self);
	return self;
}
#endif

- yourself
{
	return self;
}

- (Class)class
{
	return isa;
}

#if 0
- (OOP)oopClass
{
	return VMp->classNameToOOP(isa->name);
}

- (OOP)oop
{
	return (OOP)self;
}

- oop:(OOP)o
{
	oop = o;
	return self;
}

- (union STProxy)proxy
{
	union STProxy p;
	id test;

	test = [STObject proxy: &p using: oop];
	if (test != self)
		[self error: "Proxy object mismatch"];

	return p;
}

- (STObject *)proxyFor:(OOP)object
{
	if (subCount >= MAX_SUBPROXIES)
		[self error: "Too many proxies"];

	return [isa proxy: &subproxies[subCount++] using: object];
}
#endif

- (Method_t)methodFor:(SEL)aSel
{
	return class_get_instance_method(isa, aSel);
}

- subclassResponsibility:(SEL)aSel
{
	return [self error: "subclass should override %s", sel_get_name(aSel)];
}

- notImplemented:(SEL)aSel
{
	return [self error: "method %s not implemented", sel_get_name(aSel)];
}

- shouldNotImplement:(SEL)aSel
{
	return [self error: "%s should not implement %s", 
	             object_get_class_name(self), sel_get_name(aSel)];
}

- doesNotRecognize:(SEL)aSel
{
	return [self error: "does not recognize #%s", sel_get_name(aSel)];
}

// TODO/FIXME: -perform{with:,with:with:} will not forward to ST

- perform:(SEL)aSel
{
	IMP msg = objc_msg_lookup(self, aSel);
	if (!msg)
		return [self error: "Performance failed"];
	return (*msg)(self, aSel);
}

- perform:(SEL)aSel with:(id)anObject
{
	IMP msg = objc_msg_lookup(self, aSel);
	if (!msg)
		return [self error: "Performance failed"];
	return (*msg)(self, aSel, anObject);
}

- perform:(SEL)aSel with:(id)anObject with:(id)anotherObject
{
	IMP msg = objc_msg_lookup(self, aSel);
	if (!msg)
		return [self error: "Performance failed"];
	return (*msg)(self, aSel, anObject, anotherObject);
}

- primError:(const char *)aString
{
	// FIXME: crashes.
	//objc_verror(self, OBJC_ERR_UNKNOWN, fmt, ap);

	if (self->oop)
	{
		OOP target = VMp->nilOOP;

		if (object_is_instance(self))
			target = self->oop;//[self asOop];
		else if (object_is_class(self))
			target = VMp->classNameToOOP(isa->name);

		OOP args[1] = { VMp->stringToOOP(aString) };
		OOP selector = VMp->symbolToOOP("error:");
		VMp->nvmsgSend(target, selector, args, 1);
		[ST backtrace];
	}
	else
	{
		//OOP sfmt = VMp->stringToOOP(fmt);
		//OOP selector = VMp->stringToOOP("error:");
		// FIXME:
		[ST evaluate: "!self error: '%s'!", aString];
	}
	//vprintf(fmt, ap);

	printf("*** Aborting\n");
	exit(1);
	//abort();
	return nil;
}

- error:(const char *)aString in:(SEL)sel, ...
{
	char fmt[512], buf[512];
	va_list args;

	sprintf(fmt, "(%s %s>>#%s) %s",
			object_is_instance(self)?"Instance":"Class",
			isa->name,
			sel_get_name(sel),
			aString);

	va_start(args, sel);
	vsprintf(buf, fmt, args);
	va_end(args);

	return [self primError: buf];
}

- error:(const char *)aString, ...
{
	char fmt[512], buf[512];
	va_list args;

	sprintf(fmt, "(%s %s) %s",
			object_is_instance(self)?"Instance":"Class",
			isa->name,
			aString);

	va_start(args, aString);
	vsprintf(buf, fmt, args);
	va_end(args);

	return [self primError: buf];
}

- (retval_t)forward:(SEL)aSel :(arglist_t)argframe
{
	// send the message as smalltalk
	// the selector MUST be defined in the @interface...

	// BIG problem: sending to a smalltalk STObject that doesn't recognize...
	// will create recursion. =)

	printf("Trying forward ........... #%s\n", sel_get_name(aSel));


	//if (object_is_class(self))
	//[self error: "Unsupported: class methods (#%s)", sel_get_name(aSel)];

	const char *type = aSel->sel_types;

	if (!type)
		[self error: "No such selector defined #%s", sel_get_name(aSel)];

	//printf("%i %i ------ \n", object_is_instance(self), object_is_class(self));

	if ((object_is_instance(self) && self->oop) || object_is_class(self))
	{
		const char *retType = type;
		int numOuts = -2;
		OOP outs[16];
		char *arg;
		OOP o, dest;

		//printf("seltype:  %s for %s\n", type, sel_get_name(aSel));

		//if (!self->oop)
		//[self error: "Forwarding without OOP, check proxy"];

		while (1)
		{
			arg = method_get_next_argument(argframe, &type);
			if (!arg)
				break;

			//printf("%c arrrrrrrg\n", *type);

			if (numOuts >= 0)
			{
				switch (*type)
				{
					//case _C_ID:
					//outs[i+2].value = self->oop;
					//break;
				case _C_CHARPTR:
					//printf("done string -- nteuhaeo...\n");
					o = VMp->stringToOOP(*(char **)arg);
					break;

				case _C_INT:
				case _C_UINT:
					o = VMp->intToOOP(*(int *)arg);
					break;

				default:
					printf("++Unsupported out type '%c'\n", *type);
					o = VMp->nilOOP;
					break;
				}

				outs[numOuts] = o;
			}

			numOuts++;
		}

		//printf("++send: %s\n", sel_get_name(aSel));
		//printf("return type %c\n", retType);

		if (object_is_class(self))
			//dest = OOP_CLASS(self->oop);
			dest = VMp->classNameToOOP(class_get_class_name((Class)self));
		else
			dest = self->oop;

		o = VMp->nvmsgSend(dest, VMp->symbolToOOP(sel_get_name(aSel)), outs, numOuts);
		//o = VMp->nvmsgSend([self asOop], VMp->symbolToOOP(sel_get_name(aSel)), outs, numOuts);

		retval_t retframe = alloca(16);
		void *test;

		// UHH HUMA - this is the same stuff as -[Smalltalk performMessage:on:args:]
		// reuse pls.

		switch (*retType)
		{
		case _C_ID: {
			// FIXME

			//STObject *sto = [STObject proxy: &subproxies[subCount++] using: o];

			//STObject *sto = [self proxyFor: o];

			//STObject *sto = [alloc...];

			//STObject *sto = [ST allocObjectForOOP: o];

			// TEMP! (this is so wrong
			union STProxy p;
			Class class = objc_lookup_class([ST classNameOf: o]);
			STObject *sto = [STObject proxy: &p class: class using: o];

			memcpy(retframe, &sto, sizeof(STObject *));
			break;
		}

#if 0
		case _C_CHR: {
			char c = VMp->OOPToChar(o);
			test[0] = (void *)c;
			break;
		} case _C_UCHR: {
			unsigned char c = VMp->OOPToBool(o);
			((unsigned char)test) = c;
			break;
		}

		case _C_SHT: {
			short s = VMp->OOPToInt(o);
			((short)test) = s;
			break;
		} case _C_USHT: {
			unsigned short s = VMp->OOPToInt(o);
			((unsigned short)test) = s;
			break;
		}
#endif

		case _C_INT: {
			int n = VMp->OOPToInt(o);
			void *test = (void *)n;
			memcpy(retframe, &test, objc_sizeof_type(retType));
			break;
		} case _C_UINT: {
			unsigned int n = VMp->OOPToInt(o);
			void *test = (void *)n;
			memcpy(retframe, &test, objc_sizeof_type(retType));
			break;
		}

		case _C_FLT: {
			// FIXME: possible precision problem here as well ?

			float f = VMp->OOPToFloat(o);
			//((float)test) = f;
			//test = (void *)f;
			memcpy(retframe, &f, objc_sizeof_type(retType));
			break;
		}

		case _C_CHARPTR: {
			char *str = VMp->OOPToString(o);
			//test = (void *)str;
			memcpy(retframe, &str, objc_sizeof_type(retType));
			break;
		} case _C_PTR: {
			PTR ptr = VMp->OOPToCObject(o);
			//test = (void *)ptr;
			memcpy(retframe, &ptr, objc_sizeof_type(retType));
			break;
		}

		default:
			//test = self;
			memcpy(retframe, &self, objc_sizeof_type(retType));
			printf("++Unsupported Return type '%c'\n", *retType);
			break;
		}

		//memcpy(retframe, test, objc_sizeof_type(retType));

		return retframe;
		//return (retval_t)self;
	}

	printf("uhhhhhhh\n");
	return (retval_t)[self doesNotRecognize: aSel];
}

@end

////////////////////

@implementation UndefinedObject

- (retval_t)forward:(SEL)aSel :(arglist_t)argframe
{
	return (retval_t)[self doesNotRecognize: aSel];
}

@end
