// Copyright (c) 2004 Lyndon Tremblay. All Right Reserved.
#include <Main.h>

#define PER_TRACE

// TODO: move performance into STObject

OOP ST_PerformMessage (OOP oop, OOP omessage)
{ return [ST performMessage: omessage on: oop args: NULL]; }

OOP ST_PerformMessage1 (OOP oop, OOP omessage, OOP a)
{ return [ST performMessage: omessage on: oop args: a, NULL]; }

OOP ST_PerformMessage2 (OOP oop, OOP omessage, OOP a, OOP b)
{ return [ST performMessage: omessage on: oop args: a, b, NULL]; }

OOP ST_PerformMessage3 (OOP oop, OOP omessage, OOP a, OOP b, OOP c)
{ return [ST performMessage: omessage on: oop args: a, b, c, NULL]; }

OOP ST_PerformMessage4 (OOP oop, OOP omessage, OOP a, OOP b, OOP c, OOP d)
{ return [ST performMessage: omessage on: oop args: a, b, c, d, NULL]; }

OOP ST_PerformMessage5 (OOP oop, OOP omessage, OOP a, OOP b, OOP c, OOP d, OOP e)
{ return [ST performMessage: omessage on: oop args: a, b, c, d, e, NULL]; }

OOP ST_PerformMessage6 (OOP oop, OOP omessage, OOP a, OOP b, OOP c, OOP d, OOP e, OOP f)
{ return [ST performMessage: omessage on: oop args: a, b, c, d, e, f, NULL]; }

///////////////////

VMProxy *VMp = &gst_interpreter_proxy;
Smalltalk *ST = nil;

@implementation Smalltalk

# if 0
+ smalltalkWithArgc:(int)argc argv:(char **)argv
{
	return ST;
}

+ vm
{
	static id vm = nil;

	if (!vm)
		vm = [self alloc];// initWithArgc: argc argv: argv];

	return vm;
}
#endif

- initWithArgc:(int)argc argv:(char **)argv
{
	gst_smalltalk_args(argc, argv);
	if (gst_init_smalltalk())
		[self error: "*** Failed to init GST"];

	VMp->defineCFunc("ST_PerformMessage", ST_PerformMessage);
	VMp->defineCFunc("ST_PerformMessage1", ST_PerformMessage1);
	VMp->defineCFunc("ST_PerformMessage2", ST_PerformMessage2);
	VMp->defineCFunc("ST_PerformMessage3", ST_PerformMessage3);
	VMp->defineCFunc("ST_PerformMessage4", ST_PerformMessage4);
	VMp->defineCFunc("ST_PerformMessage5", ST_PerformMessage5);
	VMp->defineCFunc("ST_PerformMessage6", ST_PerformMessage6);

	//oop = VMp->nilOOP;   // this should be the Smalltalk dictionary ?...
	oop = NULL;

	return self;
}

- interpret
{
	gst_top_level_loop();
	return self;
}

- (OOP)evaluate:(char *)format, ...
{
	va_list args;
	char buf[1024];

	va_start(args, format);
	vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);

	return VMp->evalExpr(buf);
}

// - (OOP)exception:(char *)name signal:(char *)format, ...
// {
// 	va_list args;
// 	char buf[1024];

// 	va_start(args, format);
// 	vsnprintf(buf, sizeof(buf), format, args);
// 	va_end(args);

// 	return VMp->evalExpr(buf);
// }

- (OOP)fileIn:(char *)fileName
{
	return [self evaluate: "FileStream fileIn: '%s'!", fileName];
}

////

struct Class_guts
{
	OBJ_HEADER;

	// object

	// behavior
	OOP superClass, subClasses, methodDictionary,
		instanceSpec, instanceVariables;

	// class description

	// class

	OOP name, comment, category, environment,
		classVariables, sharedPools;
};

- (char *)classNameOf:(OOP)object
{
	/*
* [07/13/04 02:49:59 +0800]
	TODO: We need to reverse -classNameOf: and -classOf:
	 */

	BOOL validClassString (char *name) {
		if (!name || !isalpha(*name))
			return NO;
		return YES;
	}

	BOOL validClassName (char **pstr, OOP ostr) {
		if (!IS_OOP(ostr))
			return NO;

		char *name = VMp->OOPToString(ostr);

#ifdef PER_TRACE
		printf("-> validClassName: %s\n" ,name);
#endif
		
		if (!validClassString(name))
		{
			*pstr = "UndefinedObject";
			return NO;
		}
		else if (!objc_lookup_class(name))
		{
			*pstr = name;
			return NO;
		}
		else
		{
			*pstr = name;
			return YES;
		}
	}

	if (!object)
		return "Nil";

	char *className;

	// big hacks.

	if (!validClassName(&className, STGuts(Class, OOP_CLASS(object))->name))
	{
		char *next;

#ifdef PER_TRACE
		//printf("-> !valid %s (try to %p)\n", className, STGuts(Class, object)->name);
#endif

		if (!validClassName(&next, STGuts(Class, object)->name))
		{
			//[self evaluate: "SystemExceptions.NotYetImplemented signal: "
			//  "'(ObjC) No implementation of '%s' (class %s)\n'!",
			//  className, next];
			//printf("++No implementation of '%s' (class %s)\n", className, next);

			if (validClassString(className))
				return className;
			else
				return NULL;
		}
		else
			return next;
	}
	else 
		return className;

	return NULL;
}

struct Message_guts
{
	OBJ_HEADER;
	OOP selector;
	OOP args;
};

#define Message_selector(o,Type) (VMp->OOPTo##Type(STGuts(Message,o)->selector))

- (char *)selectorNameOf:(OOP)msg
{
	if (!msg)
		return "nil";

	//return STGuts(Message,msg)->selector;
	return Message_selector(msg,String);
}

- (SEL)selectorOf:(OOP)msg
{
	return sel_get_uid([self selectorNameOf: msg]);
}

struct STArgument
{
	char type;
	int offset, size;
	void *value; // late bound after lookup+parse
};

struct STMessage
{
	const char *name;
	//Class class;
	//char *targetName;

	STObject *target;
	union STProxy proxy;

	Method_t method;
	int numArgs;
	struct STArgument args[16];     // [0] is return type

	int numProxies;
	union STProxy proxies[16];
};

- (OOP)lookupMessage:(OOP)message
				   on:(OOP)object
				 into:(struct STMessage *)msg
{
	//msg->targetName = "UndefinedObject";

	if (!object)
	{
		[self error: "Not an object for #%s", [self selectorNameOf: message]];
		return NULL;
// 		return [self evaluate: "SystemExceptions.InvalidValue signal: "
// 					 "'(ObjC) Not an object for #%s'!",
// 					 [self selectorNameOf: message]];
	}

// 	if (!OOP_CLASS(object))
// 		[self error: "Class method #%s not supported", [self selectorNameOf: message]];

	BOOL IsFilerChar (char c) {
		static char *filers = "@#:cCsSiIlLqQfdbv?^*%[](){}!";
		int i = 0;

		while (filers[i])
			if (c == filers[i++])
				return YES;
		return NO;
	}

	char *ParseArgument (char *ptr, struct STArgument *arg) {
		char offsetStr[16];
		int i = 0;

		if (!*ptr || !IsFilerChar(*ptr))
			return NULL;

		arg->value = NULL;
		arg->type = *ptr++;

		// hack
		if (*ptr == '{')
		{
			int bc = 1;
			while (*ptr++)
			{
				if (*ptr == '}')
					bc--;
				if (*ptr == '{')
					bc++;

				if (!bc)
					break;
			}

			ptr++;
		}

		while (*ptr && !IsFilerChar(*ptr))
			offsetStr[i++] = *ptr++;

		offsetStr[i] = 0;
		arg->offset = atoi(offsetStr);

		return ptr;
	}

	msg->name = Message_selector(message,String);
	msg->numArgs = 0;

	SEL sel = sel_get_uid(msg->name);

#ifdef PER_TRACE
	printf("Looking up ........... #%s\n", msg->name);
#endif

	// follow hierarchy with the method

	char *targetName = [self classNameOf: object];
	Class class = objc_lookup_class(targetName);

	if (!class)
	{
		[self error: "No ObjC implementation for %s", targetName];
		return NULL;
	}

	while (class)
	{
		msg->method = class_get_instance_method(class, sel);

		// TODO: class methods =)
		if (!msg->method)
		{
#ifdef PER_TRACE
			printf("Trying Class ....%s\n", targetName);
#endif
			msg->method = class_get_class_method(class, sel);
		}

		if (msg->method)
			break;

		class = class_get_super_class(class);
	}

	if (!msg->method)
	{
		[self error: "No such selector for [%s %s], please #method or describe in @interface",
			  targetName, msg->name];
		return NULL;
	}

	msg->target = [STObject proxy: &msg->proxy class: class using: object];
	//msg->target = [self proxyFor: object];
	//msg->target = [self allocObjectForOOP: object];

	if (!msg->target)
	{
		[self error: "No ObjC response for [%s %s]", targetName, msg->name];
		return NULL;
	}


#ifdef PER_TRACE
	printf("Encoding: %s\n", msg->method->method_types);
#endif

	// parse arguments
	char *ptr = (char *)msg->method->method_types;
	int lastOffset = 0;

	while (msg->numArgs <= 16)
	{
		struct STArgument *arg = &msg->args[msg->numArgs];

		ptr = ParseArgument(ptr, arg);
		if (!ptr)
			break;

		msg->numArgs++;
		if (arg->type == _C_ID)
			msg->numProxies++;

		// temp
		//if (msg->numArgs > 2 && arg->type != _C_ID)
		//printf("++Non-object argument not yet supported...\n");

		// fixme, ignore stack size...
		arg->size = arg->offset - lastOffset;
		lastOffset = arg->offset;
		//printf("arg %c size %i offset %i\n", arg->type, arg->size, arg->offset);
	}

	// evaluate and fixup, sizes...
	for (int i=0 ; i<msg->numArgs ; i++)
	{
		struct STArgument *arg = &msg->args[msg->numArgs];
	}

	return object;
}

- (OOP)performMessage:(OOP)message on:(OOP)object args:(OOP)oarg,...
{
	struct STMessage msg;
	OOP ret;

	ret = [self lookupMessage: message on: object into: &msg];
	if (ret != object)
#if 0
		return ret;

	if (0)
#endif
	{
		[self error: "(Proxy)%s did not understand #%s\n",
			  //msg.className,
			  msg.target ? msg.target->isa->name : "UndefinedObject",
			  msg.name];

		return VMp->nilOOP;

// 		[self exception: "Smalltalk.MessageNotUnderstood"
// 			  signal: "%s did not understand #%s\n",
// 			  msg.target->isa->name, msg.name];

// 		return [self evaluate: "MessageNotUnderstood signal: "
// 					 "'(ObjC) %s did not understand #%s'!", msg.target->isa->name, msg.name];

	}

	int nargs = msg.numArgs - 3;
	void *vargs[msg.numArgs];

	memset(vargs, 0, msg.numArgs*sizeof(void*));
	//vargs[0] = ST_ProxyObject(msg.target);
	vargs[0] = msg.target;
	vargs[1] = (void *)msg.method->method_name;

	// collect arguments T_T

	va_list args;
	OOP next = oarg;
	int i = 0;

	va_start(args, oarg);
	while (next != NULL)
	{
		if (i > nargs)
		{
			printf("++%i too many arguments passed to method\n", i-nargs);
			break;
		}

		struct STArgument *arg = &msg.args[i+3];

		// FIXME: etc temp ,use self for objects
		//arg->value = oop;
		// use oarg =)

		switch (arg->type)
		{
		case _C_ID: {
			//Class c = ST_ClassOfOOP(next);
			//msg.proxies[i].object.class_pointer = c;

			Class class = objc_lookup_class([self classNameOf: next]);

			STObject *sto = [STObject proxy: (msg.proxies + i)
									  class: class
									  using: next];

			//STObject *sto = [self proxyFor: next];

			//STObject *sto = [self allocObjectForOOP: next];
			vargs[i+2] = sto;
			break;
		}

#if 0
		case _C_CHR: {
			char c = VMp->OOPToChar(next);
			vargs[i+2] = (void *)c;
			break;
		} case _C_UCHR: {
			unsigned char c = VMp->OOPToBool(next);
			vargs[i+2] = (void *)c;
			break;
		}

		case _C_SHT: {
			short s = VMp->OOPToInt(next);
			vargs[i+2] = (void *)s;
			break;
		} case _C_USHT: {
			unsigned short s = VMp->OOPToInt(next);
			vargs[i+2] = (void *)s;
			break;
		}
#endif

		case _C_INT: {
			int n = VMp->OOPToInt(next);
			vargs[i+2] = (void *)n;
			break;
		} case _C_UINT: {
			unsigned int n = VMp->OOPToInt(next);
			vargs[i+2] = (void *)n;
			break;
		}

		case _C_LNG: {
			long n = VMp->OOPToInt(next);
			vargs[i+2] = (void *)n;
			break;
		} case _C_ULNG: {
			unsigned long n = VMp->OOPToInt(next);
			vargs[i+2] = (void *)n;
			break;
		}

#if 0
		// these two may be very broken
		case _C_LNG_LNG: {
			long long n = VMp->OOPToInt(next);
			vargs[i+2] = (void *)n;
			break;
		} case _C_ULNG_LNG: {
			unsigned long long n = VMp->OOPToInt(next);
			vargs[i+2] = (void *)n;
			break;
		}
#endif

		case _C_FLT: {
		// FIXME: there is a precision problem here... investigate please
			float f = VMp->OOPToFloat(next);
			((float*)vargs)[i+2] = f;
			break;
		} 

#if 0
		case _C_DBL: {
			double d = VMp->OOPToFloat(next);
			memcpy(vargs[i+2], &d, sizeof(void *));
			break;
		}
#endif

		case _C_CHARPTR: {
			// is this safe...
			char *str = VMp->OOPToString(next);
			vargs[i+2] = (void *)str;
			break;
		} case _C_PTR: {
			PTR ptr = VMp->OOPToCObject(next);
			vargs[i+2] = (void *)ptr;
			break;
		}

		default:
			printf("**sending null for invalid argtype %c...\n", arg->type);
			vargs[i+2] = NULL;
			break;
		}

		i++;
		next = va_arg(args, OOP);
	}
	va_end(args);

	if (i < nargs)
	{
		printf("++%i too little arguments passed for [%s %s] encoded %s needing %i\n",
			   nargs-i-1, [self classNameOf: object],  msg.name, msg.method->method_types,
			   nargs);

		printf("Types:\n");
		for (i=0 ; i<msg.numArgs ; i++)
			printf("%c\n", msg.args[i].type);
	}

	IMP imp = msg.method->method_imp;
	id retval = self;

#ifdef PER_TRACE
	printf("perform: #%s\n", msg.name);
#endif

	// TODO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	// use ARGFRAME and perform: ....................
	// kthx <3 magi!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	// eek
	switch (msg.numArgs - 3)
	{
	case 0:
		retval = imp(vargs[0], vargs[1]);
		break;
	case 1:
		retval = imp(vargs[0], vargs[1], vargs[2]);
		break;
	case 2:
		retval = imp(vargs[0], vargs[1], vargs[2], vargs[3]);
		break;
	case 3:
		retval = imp(vargs[0], vargs[1], vargs[2], vargs[3], vargs[4]);
		break;
	case 4:
		retval = imp(vargs[0], vargs[1], vargs[2], vargs[3], vargs[4], vargs[5]);
		break;
	case 5:
		retval = imp(vargs[0], vargs[1], vargs[2], vargs[3], vargs[4], vargs[5], vargs[6]);
		break;
	default:
		printf("**Unsupported argument count (%i) for [%s %s] encoded %s\n",
			   msg.numArgs, [self classNameOf: object],  msg.name, msg.method->method_types);

		printf("Types:\n");
		for (i=0 ; i<msg.numArgs ; i++)
			printf("%c\n", msg.args[i].type);

		break;
	}

	//printf("rettype = %c\n", msg.args[0].type);

	switch (msg.args[0].type)
	{
	case _C_ID:
		// ouch, this won't work exactly...
		//return ((STObject *)retval)->oop;

		// temp
		return object;

	case _C_INT:
		return VMp->intToOOP((int)retval);
	case _C_UINT:
		return VMp->intToOOP((unsigned int)retval);

	case _C_FLT:
		// FIXME: there is a precision problem here... investigate please
		//return VMp->floatToOOP(999.111);
		return VMp->floatToOOP(*(float *)&retval);

	case _C_CHR:
		return VMp->charToOOP((int)retval);

	case _C_UCHR:
		// fixme, ....
		return VMp->boolToOOP((int)retval);

	case _C_PTR:
		return VMp->cObjectToOOP((id)retval);

	default:
		printf("**Unhandled return type '%c' for [%s %s]\n", msg.args[0].type,
			   msg.target->isa->name,
			   msg.name);
		return VMp->nilOOP;
	}

	//return (OOP)self;
	return oop; // FIXME.
}

- backtrace
{
	[ST evaluate: "Smalltalk backtrace!"];
	return self;
}

- forEachSubStringIn:(char *)string do:(SEL)selector on:(id)object
{
	char temp[256];
	int i, pos, len = strlen(string);

	BOOL isWhitespace (char c) {
		return (c <= 32 || c >= 127) ? YES : NO;
	}

	for (pos=0 ; pos<len ; )
	{
		i = 0;

		while (string[pos] && isWhitespace(string[pos]))
			pos++;

		while (string[pos] && !isWhitespace(string[pos]))
			temp[i++] = string[pos++];

		if (!i)
			break;

		temp[i] = 0;
		[object perform: selector with: (id)temp];
	}

	return self;
}

@end

